"""
This module contains a base class for an EPANET-MSX control environment --
i.e. controlling the injection and reaction of one or multiple species in an
EPANET-MSX scenario (no control over pumps, valves, etc.).
"""
import os
import uuid
from typing import Optional, Any
import numpy as np
from epyt_flow.simulation import ScenarioConfig, ScenarioSimulator
from epyt_flow.utils import get_temp_folder
from gymnasium.spaces import Dict
from gymnasium.spaces.utils import flatten_space
from gymnasium import Env

from ..actions.quality_actions import SpeciesInjectionAction
from .rl_env import RlEnv


class EpanetMsxControlEnv(RlEnv):
    """
    Base class for advanced quality control scenarios -- i.e. EPANET-MSX control scenarios.

    Parameters
    ----------
    scenario_config : `epyt_flow.simulation.ScenarioConfig <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.html#epyt_flow.simulation.scenario_config.ScenarioConfig>`_
        Configuration of the scenario.
    action_space : list[:class:`~epyt_control.actions.quality_actions.SpeciesInjectionAction`]
        The action spaces (i.e. list of species injections) that have to be controlled by the agent.
    rerun_hydraulics_when_reset : `bool`
        If True, the hydraulic simulation is going to be re-run when the environment is reset,
        otherwise the hydraulics from the initial run are re-used and the scenario will
        also not be reloaded -- i.e. reload_scenario_when_reset=False.
    """
    def __init__(self, scenario_config: ScenarioConfig,
                 action_space: list[SpeciesInjectionAction],
                 rerun_hydraulics_when_reset: bool = True, **kwds):
        if not isinstance(action_space, list):
            raise TypeError("'action_space' must be an instance of " +
                            "`list[SpeciesInjectionActionSpace]` " +
                            f"but not of '{type(action_space)}'")
        if any(not isinstance(action_desc, SpeciesInjectionAction)
               for action_desc in action_space):
            raise TypeError("All items in 'action_space' must be an instance of " +
                            "'epyt_control.actions.quality_actions.SpeciesInjectionAction'")
        if len(action_space) == 0:
            raise ValueError("Empty action space")
        if not isinstance(rerun_hydraulics_when_reset, bool):
            raise TypeError("'rerun_hydraulics_when_reset' must be an instance of 'bool' " +
                            f"but not of '{type(rerun_hydraulics_when_reset)}'")
        if "reload_scenario_when_reset" in kwds:
            if kwds["reload_scenario_when_reset"] is True and rerun_hydraulics_when_reset is False:
                raise ValueError("'rerun_hydraulics_when_reset' must be True if 'reload_scenario_when_reset=True'")
        else:
            if rerun_hydraulics_when_reset is False:
                kwds["reload_scenario_when_reset"] = False

        self._rerun_hydraulics_when_reset = rerun_hydraulics_when_reset
        self._hyd_export = os.path.join(get_temp_folder(),
                                        f"epytcontrol_env_MSX_{uuid.uuid4()}.hyd")
        gym_action_space = flatten_space(Dict({f"{action_space.species_id}-{action_space.node_id}":
                                               action_space.to_gym_action_space()
                                               for action_space in action_space}))

        super().__init__(scenario_config=scenario_config, gym_action_space=gym_action_space,
                         action_space=action_space, **kwds)

    def reset(self, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
              ) -> tuple[np.ndarray, dict]:
        Env.reset(self, seed=seed)

        if self._rerun_hydraulics_when_reset is True:
            scada_data = super().reset()
        else:
            if self._scenario_sim is None or self._reload_scenario_when_reset:
                self._scenario_sim = ScenarioSimulator(
                    scenario_config=self._scenario_config)

                # Run hydraulic simulation first
                sim = self._scenario_sim.run_hydraulic_simulation
                self._hydraulic_scada_data = sim(hyd_export=self._hyd_export)
            else:
                # Abort current simulation if any is runing
                try:
                    next(self._sim_generator)
                    self._sim_generator.send(True)
                except StopIteration:
                    pass

            # Run advanced quality analysis (EPANET-MSX) on top of the computed hydraulics
            gen = self._scenario_sim.run_advanced_quality_simulation_as_generator
            self._sim_generator = gen(self._hyd_export, support_abort=True)

            scada_data = self._next_sim_itr()

        r = scada_data
        if isinstance(r, tuple):
            r, _ = r
        r = self._get_observation(r)

        return r, {"scada_data": scada_data}


AdvancedQualityControlEnv = EpanetMsxControlEnv


class MultiConfigEpanetMsxControlEnv(EpanetMsxControlEnv):
    """
    Base class for advanced quality control scenarios (i.e. EPANET-MSX control scenarios) that can
    handle multiple scenario configurations -- those scenarios are utilized in a round-robin
    scheduling scheme (i.e. autorest=True).

    Parameters
    ----------
    scenario_configs : list[`epyt_flow.simulation.ScenarioConfig <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.html#epyt_flow.simulation.scenario_config.ScenarioConfig>`_]
        Configuration of the scenario.
    action_space : list[:class:`~epyt_control.actions.quality_actions.SpeciesInjectionAction`]
        The action spaces (i.e. list of species injections) that have to be controlled by the agent.
    rerun_hydraulics_when_reset : `bool`
        If True, the hydraulic simulation is going to be re-run when the environment is reset,
        otherwise the hydraulics from the initial run are re-used and the scenario will
        also not be reloaded -- i.e. reload_scenario_when_reset=False.
    """
    def __init__(self, scenario_configs: list[ScenarioConfig],
                 action_space: list[SpeciesInjectionAction],
                 rerun_hydraulics_when_reset: bool = True, **kwds):
        if not isinstance(scenario_configs, list):
            raise TypeError("'scenario_configs' must be an instance of " +
                            "epyt_flow.simulation.ScenarioConfig but " +
                            f"not of '{type(scenario_configs)}'")
        if any(not isinstance(scenario_config, ScenarioConfig)
               for scenario_config in scenario_configs):
            raise TypeError("All items in 'scenario_config' must be instances of " +
                            "epyt_flow.simulation.ScenarioConfig")

        self._scenario_configs = scenario_configs
        self._scenario_sims = [None] * len(scenario_configs)
        self._hyd_exports = [os.path.join(get_temp_folder(),
                                          f"epytcontrol_env_MSX_{uuid.uuid4()}.hyd")
                             for _ in range(len(scenario_configs))]
        self._current_scenario_idx = 0

        super().__init__(self._scenario_configs[self._current_scenario_idx],
                         action_space, rerun_hydraulics_when_reset,
                         autoreset=True, **kwds)
        self._hyd_export = self._hyd_exports[self._current_scenario_idx]

    def reset(self, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
              ) -> tuple[np.ndarray, dict]:
        # Back up current simulation
        self._scenario_sims[self._current_scenario_idx] = self._scenario_sim

        # Move on to next scenario
        self._current_scenario_idx = self._current_scenario_idx + 1 % len(self._scenario_configs)
        self._scenario_config = self._scenario_configs[self._current_scenario_idx]
        self._scenario_sim = self._scenario_sims[self._current_scenario_idx]
        self._hyd_export = self._hyd_exports[self._current_scenario_idx]

        return super().reset(seed, options)


MultiConfigAdvancedQualityControlEnv = MultiConfigEpanetMsxControlEnv



"""
This module contains classes for modeling actuator state action spaces --
i.e. controlling pump and valve states.
"""
from copy import deepcopy
from epyt_flow.gym import ScenarioControlEnv
from epyt_flow.simulation.events import ActuatorConstants
from gymnasium.spaces import Space, Discrete

from .actions import Action


class ActuatorStateAction(Action):
    """
    Base class of an actuator state action.

    Parameters
    ----------
    state_space : `list[int]`
        List of possible actions that can be taken by the agent.
    """
    def __init__(self, state_space: list[int], **kwds):
        if not isinstance(state_space, list):
            raise TypeError("'state_space' must be an instance of 'list[int]' " +
                            f"but not of '{type(state_space)}'")
        if any(not isinstance(state, int) for state in state_space):
            raise TypeError("All states in 'state_space' must be an instance of 'int'")
        self._action_space = state_space

        super().__init__(**kwds)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._action_space == other.action_space

    def __str__(self) -> str:
        return super().__str__() + f"Action space: {self._action_space}"

    @property
    def action_space(self) -> list[int]:
        """
        Returns the list of possible actions that an agent can take.

        Returns
        -------
        `list[int]`
            List of possible actions.
        """
        return deepcopy(self._action_space)

    def to_gym_action_space(self) -> Space:
        return Discrete(len(self._action_space))


class ValveStateAction(ActuatorStateAction):
    """
    Action for controlling the state of a valve.

    Parameters
    ----------
    valve_id : `str`
        ID of the valve.
    """
    def __init__(self, valve_id: str, **kwds):
        self._valve_id = valve_id

        super().__init__(state_space=[ActuatorConstants.EN_OPEN, ActuatorConstants.EN_CLOSED],
                         **kwds)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.valve_id == other.valve_id

    def __str__(self) -> str:
        return super().__str__() + f"Valve ID: {self._valve_id}"

    @property
    def valve_id(self) -> str:
        """
        Returns the ID of the valve.

        Returns
        -------
        `str`
            ID of the valve.
        """
        return self._valve_id

    def apply(self, env: ScenarioControlEnv, action_value: int) -> None:
        env.set_valve_status(self._valve_id, action_value)


class PumpStateAction(ActuatorStateAction):
    """
    Action for controling the state of a pump.

    Parameters
    ----------
    pump_id : `str`
        ID of the pump.
    """
    def __init__(self, pump_id: str, **kwds):
        self._pump_id = pump_id

        super().__init__(state_space=[ActuatorConstants.EN_OPEN, ActuatorConstants.EN_CLOSED],
                         **kwds)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._pump_id == other.pump_id

    def __str__(self) -> str:
        return super().__str__() + f"Pump ID: {self._pump_id}"

    @property
    def pump_id(self) -> str:
        """
        Return the ID of the pump.

        Returns
        -------
        `str`
            ID of the pump.
        """
        return self._pump_id

    def apply(self, env: ScenarioControlEnv, action_value) -> None:
        env.set_pump_status(self._pump_id, action_value)



"""
This module contains a base class for actions.
"""
from abc import abstractmethod
from gymnasium.spaces import Space
from epyt_flow.gym import ScenarioControlEnv


class Action():
    """
    Base class for actions.
    """
    @abstractmethod
    def to_gym_action_space(self) -> Space:
        """
        Converts this action to a
        `gymnasium.spaces.Space <https://gymnasium.farama.org/api/spaces/#gymnasium.spaces.Space>`_
        instance.

        Returns
        -------
        `gymnasium.spaces.Space <https://gymnasium.farama.org/api/spaces/#gymnasium.spaces.Space>`_
            gymnasium.spaces.Space instance.
        """
        raise NotImplementedError()

    @abstractmethod
    def apply(self, env: ScenarioControlEnv, action_value) -> None:
        """
        Applies a given action (from the action space) in a given environment.

        Parameters
        ----------
        env : `epyt_flow.gym.ScenarioControlEnv <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.gym.html#epyt_flow.gym.scenario_control_env.ScenarioControlEnv>`_
            The environment.
        action_value: Any
            The action.
        """
        raise NotImplementedError()



"""
This module contains an implementation of a PID controller and a
learner (i.e. tuning method) based on an evolutionary algorithm.
"""
from typing import Optional
from copy import deepcopy
import numpy as np

from ..envs.rl_env import RlEnv


class PidController():
    """
    Implementation of a Proportional-Integral-Derivative (PID) controller.

    Parameters
    ----------
    proportional_gain : `float`
        Proportional gain coefficient.
    integral_gain : `float`
        Integral gain coefficient.
    derivative_gain : `float`
        Derivative gain coefficient.
    target_value : `float`
        Target value (observed system state) which the controller is supposed to reach.
    action_lower_bound : `float`, optional
        Lower bound of the computed action.
        Smaller control outputs will be clipped.

        The default is None.
    action_upper_bound : `float`, optional
        Upper bound of the computed action.
        Control outputs exceeding this upper bound will be clipped.

        The default is None.
    """
    def __init__(self, proportional_gain: float, integral_gain: float, derivative_gain: float,
                 target_value: float, action_lower_bound: Optional[float] = None,
                 action_upper_bound: Optional[float] = None):
        if not isinstance(proportional_gain, float):
            raise TypeError("'proportional_gain' must be an instance of 'float' " +
                            f"but not of '{type(proportional_gain)}'")
        if not isinstance(integral_gain, float):
            raise TypeError("'integral_gain' must be an instance of 'float' " +
                            f"but not of '{type(integral_gain)}'")
        if not isinstance(derivative_gain, float):
            raise TypeError("'derivative_gain' must be an instance of 'float' " +
                            f"but not of '{type(derivative_gain)}'")
        if not isinstance(target_value, float):
            raise TypeError("'target_value' must be an instance of 'float' " +
                            f"but not of '{type(target_value)}'")
        if action_lower_bound is not None:
            if not isinstance(action_lower_bound, float):
                raise TypeError("'action_lower_bound' must be an instance of 'float' " +
                                f"but not of '{type(action_lower_bound)}'")
        if action_upper_bound is not None:
            if not isinstance(action_upper_bound, float):
                raise TypeError("'action_upper_bound' must be an instance of 'float' " +
                                f"but not of '{type(action_upper_bound)}'")
        if action_upper_bound is not None and action_lower_bound is not None:
            if action_lower_bound >= action_upper_bound:
                raise ValueError("'action_lower_bound' must be smaller than 'action_upper_bound'")

        self._proportional_gain = proportional_gain
        self._derivative_gain = derivative_gain
        self._integral_gain = integral_gain
        self._target_value = target_value
        self._action_lower_bound = action_lower_bound
        self._action_upper_bound = action_upper_bound

        self._last_error = 0
        self._integral = 0

    def __str__(self) -> str:
        return f"proportional_gain: {self._proportional_gain} " + \
            f"derivative_gain: {self._derivative_gain} integral_gain: {self._integral_gain} " + \
            f"target_value: {self._target_value} action_lower_bound: {self._action_lower_bound}" + \
            f" action_upper_bound: {self._action_upper_bound}"

    def __eq__(self, other) -> bool:
        if not isinstance(other, PidController):
            raise TypeError(f"Can not compare 'PidController' to '{type(other)}'")

        return self._proportional_gain == other.proportional_gain and \
            self._derivative_gain == other.derivative_gain and \
            self._integral_gain == other.integral_gain and \
            self._target_value == other.target_value and \
            self._action_lower_bound == other.action_lower_bound and \
            self._action_upper_bound == other.action_upper_bound

    @property
    def proportional_gain(self) -> float:
        """
        Returns the proportional gain coefficient.

        Returns
        -------
        `float`
            Proportional gain coefficient.
        """
        return self._proportional_gain

    @property
    def integral_gain(self) -> float:
        """
        Returns the integral gain coefficient.

        Returns
        -------
        `float`
            Integral gain coefficient.
        """
        return self._integral_gain

    @property
    def derivative_gain(self) -> float:
        """
        Returns the derivative gain coefficient.

        Returns
        -------
        `float`
            Derivative gain coefficient.
        """
        return self._derivative_gain

    @property
    def target_value(self) -> float:
        """
        Returns the target value (observed system state) which the controller is supposed to reach.

        Returns
        -------
        `float`
            Target value (system state).
        """
        return self._target_value

    @property
    def action_lower_bound(self) -> float:
        """
        Lower bound of the computed action.
        Smaller control outputs will be clipped.

        Returns
        -------
        `float`
            Lower bound of the computed action.
        """
        return self._action_lower_bound

    @property
    def action_upper_bound(self) -> float:
        """
        Upper bound of the computed action.
        Control outputs exceeding this upper bound will be clipped.

        Returns
        -------
        `float`
            Upper bound of the computed action.
        """
        return self._action_upper_bound

    def step(self, cur_value: float) -> float:
        """
        Computes the current/next control action/signal based on the
        given observation (i.e. system state).

        Parameters
        ----------
        cur_value : `float`
            Current observation (i.e. system state).

        Returns
        -------
        `float`
            Computed action -- i.e. control signal.
        """
        error = self._target_value - cur_value
        self._integral += self._integral_gain * error

        action = self._proportional_gain * error + self._integral_gain * self._integral + \
            (self._derivative_gain * (error - self._last_error))
        if np.isnan(action):
            action = 0

        # Clip if action is outside of bounds
        if self._action_lower_bound is not None:
            action = max(action, self._action_lower_bound)
        if self._action_upper_bound is not None:
            action = min(action, self._action_upper_bound)

        self._last_error = error
        return action



"""
This module contains classes for modeling chemical/species injection action spaces
-- i.e. a chemical (EPANET) or species (EPANET-MSX) injection that has to be
controlled by the agent.
"""
from typing import Optional
from epyt_flow.gym import ScenarioControlEnv
from epyt_flow.simulation import ToolkitConstants
from gymnasium.spaces import Space, Box

from .actions import Action


class ChemicalInjectionAction(Action):
    """
    Action for controlling the injection of a chemical -- only for EPANET control scenarios.

    Parameters
    ----------
    node_id : `str`
        ID of the node at which the injection is going to happen.
    pattern_id : `str`
        ID of the pattern that is used for the injection.
    source_type_id : `int`
        Type of the injection source -- must be one of
        the following EPANET toolkit constants:

            - EN_CONCEN     = 0
            - EN_MASS       = 1
            - EN_SETPOINT   = 2
            - EN_FLOWPACED  = 3

        Description:

            - E_CONCEN Sets the concentration of external inflow entering a node
            - EN_MASS Injects a given mass/minute into a node
            - EN_SETPOINT Sets the concentration leaving a node to a given value
            - EN_FLOWPACED Adds a given value to the concentration leaving a node
    upper_bound : `float`, optional
        Upper bound on the amount that can be injected by the agent.
        If None, there is no upper bound on the amount.

        The default is None.
    """
    def __init__(self, node_id: str, pattern_id: str,
                 source_type_id: int, upper_bound: Optional[float] = None,
                 **kwds):
        if not isinstance(node_id, str):
            raise TypeError(f"'node_id' must be an instance of 'str' but not of '{type(node_id)}'")
        if not isinstance(pattern_id, str):
            raise TypeError("'pattern_id' must be an instance of 'str' " +
                            f"but not of '{type(pattern_id)}'")
        if not isinstance(source_type_id, int):
            raise TypeError("'source_type_id' must be an instance of 'int' " +
                            f"but not of '{type(source_type_id)}'")
        if source_type_id not in [ToolkitConstants.EN_MASS, ToolkitConstants.EN_CONCEN,
                                  ToolkitConstants.EN_SETPOINT, ToolkitConstants.EN_FLOWPACED]:
            raise ValueError("Invalid 'source_type_id'")
        if upper_bound is not None:
            if not isinstance(upper_bound, float):
                raise TypeError("'upper_bound' must be an instance of 'float' " +
                                f"but not of '{type(upper_bound)}'")
            if upper_bound <= 0:
                raise ValueError("'upper_bound' must be positive")

        self._node_id = node_id
        self._pattern_id = pattern_id
        self._source_type_id = source_type_id
        self._upper_bound = upper_bound

        super().__init__(**kwds)

    @property
    def node_id(self) -> str:
        """
        Return the ID of the node at which the injection is going to happen.

        Returns
        -------
        `str`
            ID of the node.
        """
        return self._node_id

    @property
    def pattern_id(self) -> str:
        """
        Returns the ID of the pattern that is used for the injection.

        Returns
        -------
        `str`
            ID of the pattern.
        """
        return self._pattern_id

    @property
    def source_type_id(self) -> int:
        """
        Returns the type (i.e. ID) of the injection source -- will be one of
        the following EPANET toolkit constants:

            - EN_CONCEN     = 0
            - EN_MASS       = 1
            - EN_SETPOINT   = 2
            - EN_FLOWPACED  = 3

        Description:

            - E_CONCEN Sets the concentration of external inflow entering a node
            - EN_MASS Injects a given mass/minute into a node
            - EN_SETPOINT Sets the concentration leaving a node to a given value
            - EN_FLOWPACED Adds a given value to the concentration leaving a node

        Returns
        -------
        `int`
            Type (ID) of the injection source.
        """
        return self._source_type_id

    @property
    def upper_bound(self) -> float:
        """
        Returns the upper bound on the amount that can be injected by the agent.

        Returns
        -------
        `float`
            Upper bound.
        """
        return self._upper_bound

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._node_id == other.node_id and \
            self._pattern_id == other.pattern_id and \
            self._source_type_id == other.source_type_id and \
            self._upper_bound == other.upper_bound

    def __str__(self) -> str:
        return f"Node ID: {self._node_id} " +\
            f"Pattern ID: {self._pattern_id} Source type ID: {self._source_type_id} " +\
            f"Upper bound: {self._upper_bound}"

    def to_gym_action_space(self) -> Space:
        return Box(low=0, high=self._upper_bound if self._upper_bound is not None
                   else float("inf"))

    def apply(self, env: ScenarioControlEnv, action_value: float) -> None:
        env.set_node_quality_source_value(self._node_id, self._pattern_id, action_value)


class SpeciesInjectionAction(ChemicalInjectionAction):
    """
    Action for controlling the injection of a given species --
    only for EPANET-MSX control scenarios.

    Parameters
    ----------
    species_id : `str`
        ID of the species that is going to be injected by the agent.
    node_id : `str`
        ID of the node at which the injection is going to happen.
    pattern_id : `str`
        ID of the pattern that is used for the injection.
    source_type_id : `int`
        Types of the injection source -- must be one of
        the following EPANET toolkit constants:

            - EN_CONCEN     = 0
            - EN_MASS       = 1
            - EN_SETPOINT   = 2
            - EN_FLOWPACED  = 3

        Description:

            - E_CONCEN Sets the concentration of external inflow entering a node
            - EN_MASS Injects a given mass/minute into a node
            - EN_SETPOINT Sets the concentration leaving a node to a given value
            - EN_FLOWPACED Adds a given value to the concentration leaving a node
    upper_bound : `float`, optional
        Upper bound on the amount that can be injected by the agent.
        If None, there is no upper bound on the amount.

        The default is None.
    """
    def __init__(self, species_id: str, node_id: str, pattern_id: str,
                 source_type_id: int, upper_bound: Optional[float] = None,
                 **kwds):
        if not isinstance(species_id, str):
            raise TypeError("'species_id' must be an instance of 'str' " +
                            f"but not of '{type(species_id)}'")

        self._species_id = species_id

        super().__init__(node_id=node_id, pattern_id=pattern_id,
                         source_type_id=source_type_id, upper_bound=upper_bound, **kwds)

    @property
    def species_id(self) -> str:
        """
        Returns the ID of the species that is going to be injected.

        Returns
        -------
        `str`
            ID of the species.
        """
        return self._species_id

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._species_id == other.species_id

    def __str__(self) -> str:
        return super().__str__() + f"Species ID: {self._species_id}"

    def apply(self, env: ScenarioControlEnv, action_value: float) -> None:
        env.set_node_species_source_value(self._species_id, self._node_id, self._source_type_id,
                                          self._pattern_id, action_value)



"""
This module contains a base class for EPANET control environments --
i.e. controlling hydraulic actuators such as pumps and valves or single chemical (no EPANET-MSX support!).
"""
from typing import Optional, Any
import numpy as np
from epyt_flow.simulation import ScenarioConfig
from gymnasium.spaces import Dict
from gymnasium.spaces.utils import flatten_space

from .rl_env import RlEnv
from ..actions.pump_speed_actions import PumpSpeedAction
from ..actions.quality_actions import ChemicalInjectionAction
from ..actions.actuator_state_actions import PumpStateAction, ValveStateAction


class EpanetControlEnv(RlEnv):
    """
    Base class for hydraulic control environments
    (incl. basic quality that can be simulated with EPANET only).

    Parameters
    ----------
    scenario_config : `epyt_flow.simulation.ScenarioConfig <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.html#epyt_flow.simulation.scenario_config.ScenarioConfig>`_
        Configuration of the scenario.
    pumps_speed_actions : list[:class:`~epyt_control.actions.pump_speed_actions.PumpSpeedAction`], optional
        List of pumps where the speed has to be controlled.

        The default is None.
    pumps_state_actions : list[:class:`~epyt_control.actions.actuator_state_actions.PumpStateAction`], optional
        Lisst of pumps where the state has to be controlled.

        The default is None.
    valves_state_actions : list[:class:`~epyt_control.actions.actuator_state_actions.ValveStateAction`], optional
        List of valves that have to be controlled.

        The default is None.
    chemical_injection_actions : list[:class:`~epyt_control.actions.quality_actions.ChemicalInjectionAction`], optional
        List chemical injection actions -- i.e. places in the network where the
        injection of the chemical has to be controlled.

        The default is None.
    """
    def __init__(self, scenario_config: ScenarioConfig,
                 pumps_speed_actions: Optional[list[PumpSpeedAction]] = None,
                 pumps_state_actions: Optional[list[PumpStateAction]] = None,
                 valves_state_actions: Optional[list[ValveStateAction]] = None,
                 chemical_injection_actions: Optional[list[ChemicalInjectionAction]] = None,
                 **kwds):
        if pumps_speed_actions is not None:
            if not isinstance(pumps_speed_actions, list):
                raise TypeError("'pumps_speed_actions' must be an instance of " +
                                "'list[PumpSpeedAction]' but not of " +
                                f"'{type(pumps_speed_actions)}'")
            if any(not isinstance(pump_speed_action, PumpSpeedAction)
                   for pump_speed_action in pumps_speed_actions):
                raise TypeError("All items in 'pumps_speed_actions' must be an instance of " +
                                "'PumpSpeedAction'")
        if pumps_state_actions is not None:
            if not isinstance(pumps_state_actions, list):
                raise TypeError("'pumps_state_actions' must be an instance of " +
                                "'list[PumpStateAction]' but not of " +
                                f"'{type(pumps_state_actions)}'")
            if any(not isinstance(pump_state_action, PumpStateAction)
                   for pump_state_action in pumps_state_actions):
                raise TypeError("All items in 'pumps_state_actions' must be an instance of " +
                                "'PumpStateAction'")
        if valves_state_actions is not None:
            if not isinstance(valves_state_actions, list):
                raise TypeError("'valves_state_actions' must be an instance of " +
                                "'list[ValveAction]' but not of " +
                                f"'{type(valves_state_actions)}'")
            if any(not isinstance(valve_state_action, ValveStateAction)
                   for valve_state_action in valves_state_actions):
                raise TypeError("All items in 'valves_state_actions' must " +
                                "be an instance of 'ValveStateAction'")
        if chemical_injection_actions is not None:
            if not isinstance(chemical_injection_actions, list):
                raise TypeError("'chemical_injection_actions' must be an instance of " +
                                "'list[ChemicalInjectionAction]' but not of " +
                                f"'{type(chemical_injection_actions)}'")
            if any(not isinstance(chemical_injection_action, ChemicalInjectionAction)
                   for chemical_injection_action in chemical_injection_actions):
                raise TypeError("All items in 'chemical_injection_actions' " +
                                "must be an instance of 'ChemicalInjectionAction'")

        self._pumps_speed_actions = pumps_speed_actions
        self._pumps_state_actions = pumps_state_actions
        self._valves_state_actions = valves_state_actions
        self._chemical_injection_actions = chemical_injection_actions

        action_space = {}
        my_actions = []
        if self._pumps_speed_actions is not None:
            my_actions += self._pumps_speed_actions
            action_space |= {f"{action_space.pump_id}-speed": action_space.to_gym_action_space()
                             for action_space in self._pumps_speed_actions}
        if self._pumps_state_actions is not None:
            my_actions += self._pumps_state_actions
            action_space |= {f"{action_space.pump_id}-state": action_space.to_gym_action_space()
                             for action_space in self._pumps_state_actions}
        if self._valves_state_actions is not None:
            my_actions += self._valves_state_actions
            action_space |= {f"{action_space.valve_id}-state": action_space.to_gym_action_space()
                             for action_space in self._valves_state_actions}
        if self._valves_state_actions is not None:
            my_actions += self._valves_state_actions
            action_space |= {f"{action_space.valve_id}-state": action_space.to_gym_action_space()
                             for action_space in self._valves_state_actions}
        if self._chemical_injection_actions is not None:
            my_actions += self._chemical_injection_actions
            action_space |= {f"{action_space.node_id}-chem": action_space.to_gym_action_space()
                             for action_space in self._chemical_injection_actions}

        gym_action_space = flatten_space(Dict(action_space))

        super().__init__(scenario_config=scenario_config, gym_action_space=gym_action_space,
                         action_space=my_actions, **kwds)


HydraulicControlEnv = EpanetControlEnv


class MultiConfigEpanetControlEnv(EpanetControlEnv):
    """
    Base class for hydraulic control environments (incl. basic quality that can be simulated
    with EPANET only) that can handle multiple scenario configurations -- those scenarios are
    utilized in a round-robin scheduling scheme (i.e. autorest=True).

    Parameters
    ----------
    scenario_configs : list[`epyt_flow.simulation.ScenarioConfig <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.html#epyt_flow.simulation.scenario_config.ScenarioConfig>`_]
        List of all scenario configurations that are used in this environment.
    pumps_speed_actions : list[:class:`~epyt_control.actions.pump_speed_actions.PumpSpeedAction`], optional
        List of pumps where the speed has to be controlled.

        The default is None.
    pumps_state_actions : list[:class:`~epyt_control.actions.actuator_state_actions.PumpStateAction`], optional
        Lisst of pumps where the state has to be controlled.

        The default is None.
    valves_state_actions : list[:class:`~epyt_control.actions.actuator_state_actions.ValveStateAction`], optional
        List of valves that have to be controlled.

        The default is None.
    chemical_injection_actions : list[:class:`~epyt_control.actions.quality_actions.ChemicalInjectionAction`], optional
        List chemical injection actions -- i.e. places in the network where the
        injection of the chemical has to be controlled.

        The default is None.
    """
    def __init__(self, scenario_configs: list[ScenarioConfig],
                 pumps_speed_actions: Optional[list[PumpSpeedAction]] = None,
                 pumps_state_actions: Optional[list[PumpStateAction]] = None,
                 valves_state_actions: Optional[list[ValveStateAction]] = None,
                 chemical_injection_actions: Optional[list[ChemicalInjectionAction]] = None,
                 reload_scenario_when_reset: bool = True):
        if not isinstance(scenario_configs, list):
            raise TypeError("'scenario_configs' must be an instance of " +
                            "epyt_flow.simulation.ScenarioConfig but " +
                            f"not of '{type(scenario_configs)}'")
        if any(not isinstance(scenario_config, ScenarioConfig)
               for scenario_config in scenario_configs):
            raise TypeError("All items in 'scenario_config' must be instances of " +
                            "epyt_flow.simulation.ScenarioConfig")

        self._scenario_configs = scenario_configs
        self._scenario_sims = [None] * len(scenario_configs)
        self._current_scenario_idx = 0

        super().__init__(self._scenario_configs[self._current_scenario_idx], pumps_speed_actions,
                         pumps_state_actions, valves_state_actions, chemical_injection_actions,
                         autoreset=True,
                         reload_scenario_when_reset=reload_scenario_when_reset)

    def reset(self, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
              ) -> tuple[np.ndarray, dict]:
        # Back up current simulation
        self._scenario_sims[self._current_scenario_idx] = self._scenario_sim

        # Move on to next scenario
        self._current_scenario_idx = self._current_scenario_idx + 1 % len(self._scenario_configs)
        self._scenario_config = self._scenario_configs[self._current_scenario_idx]
        self._scenario_sim = self._scenario_sims[self._current_scenario_idx]

        return super().reset(seed, options)


MultiConfigHydraulicControlEnv = MultiConfigEpanetControlEnv



"""
This module contains a base class for reinforcement learning (RL) environments.
"""
from abc import abstractmethod
import os
import uuid
from copy import deepcopy
from typing import Optional, Any, Union
import numpy as np
from epyt_flow.simulation import ScadaData, ScenarioConfig, ScenarioSimulator
from epyt_flow.gym import ScenarioControlEnv
from epyt_flow.utils import get_temp_folder
from gymnasium import Env
from gymnasium.spaces import Space, Box, Discrete, Tuple
from gymnasium.spaces.utils import flatten_space

from ..actions import Action


class RlEnv(ScenarioControlEnv, Env):
    """
    Base class for reinforcement learning environments.

    Parameters
    ----------
    scenario_config : `epyt_flow.simulation.ScenarioConfig <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.html#epyt_flow.simulation.scenario_config.ScenarioConfig>`_
        Config of the scenario.
    gym_action_space : `gymnasium.spaces.Space <https://gymnasium.farama.org/api/spaces/#gymnasium.spaces.Space>`_
        Gymnasium action space.
    action_space : list[:class:`~epyt_control.actions.actions.Action`]
        List of all action spaces -- one space for each element that can be controlled by the agent.
    reload_scenario_when_reset : `bool`, optional
        If True, the scenario (incl. the .inp and .msx file) is reloaded from the hard disk.
        If False, only the simulation is reset.

        The default is True.
    """
    def __init__(self, scenario_config: ScenarioConfig, gym_action_space: Space,
                 action_space: list[Action], reload_scenario_when_reset: bool = True,
                 **kwds):
        if not isinstance(gym_action_space, Space):
            raise TypeError("'gym_action_space' must be an instance of 'gymnasium.spaces.Space' " +
                            f"but not of '{type(gym_action_space)}'")
        if not isinstance(action_space, list):
            raise TypeError("'action_spaces' must be an instance of " +
                            "'list[epyt_control.actions.Action]' " +
                            f"but not of '{type(action_space)}'")
        if any(not isinstance(a_s, Action) for a_s in action_space):
            raise TypeError("Every item in 'action_spaces' must be an instance of " +
                            "'epyt_control.actions.Action'")
        if not isinstance(reload_scenario_when_reset, bool):
            raise TypeError("'reload_scenario_when_reset' must be an instance of 'bool' " +
                            f"but not of '{type(reload_scenario_when_reset)}'")

        super().__init__(scenario_config=scenario_config, **kwds)

        self._observation_space = self._get_observation_space()
        self._action_space = action_space
        self._gym_action_space = gym_action_space
        self._reload_scenario_when_reset = reload_scenario_when_reset

    def _get_observation_space(self) -> Space:
        obs_space = []
        sensor_config = self._scenario_config.sensor_config

        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.pressure_sensors)
        obs_space += [Box(low=float("-inf"), high=float("inf"))] * len(sensor_config.flow_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.demand_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.quality_node_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.quality_link_sensors)
        obs_space += [Discrete(2, start=2)] * len(sensor_config.valve_state_sensors)
        obs_space += [Discrete(2, start=2)] * len(sensor_config.pump_state_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.pump_efficiency_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * \
            len(sensor_config.pump_energyconsumption_sensors)
        obs_space += [Box(low=0, high=float("inf"))] * len(sensor_config.tank_volume_sensors)
        for species_id in sensor_config.surface_species_sensors:
            obs_space += [Box(low=0, high=float("inf"))] * \
                len(sensor_config.surface_species_sensors[species_id])
        for species_id in sensor_config.bulk_species_node_sensors:
            obs_space += [Box(low=0, high=float("inf"))] * \
                len(sensor_config.bulk_species_node_sensors[species_id])
        for species_id in sensor_config.bulk_species_link_sensors:
            obs_space += [Box(low=0, high=float("inf"))] * \
                len(sensor_config.bulk_species_link_sensors[species_id])

        return flatten_space(Tuple(obs_space))

    @property
    def observation_space(self) -> Space:
        """
        Returns the observation space of this environment.

        Returns
        -------
        `gymnasium.spaces.Space <https://gymnasium.farama.org/api/spaces/#gymnasium.spaces.Space>`_
            Gymnasium (observation) space instance.
        """
        return self._observation_space

    @property
    def action_space(self) -> Space:
        """
        Returns the action space of this environment.

        Returns
        -------
        `gymnasium.spaces.Space <https://gymnasium.farama.org/api/spaces/#gymnasium.spaces.Space>`_
            Gymnasium (action) space instance.
        """
        return self._gym_action_space

    def _next_sim_itr(self) -> Union[tuple[ScadaData, bool], ScadaData]:
        try:
            next(self._sim_generator)
            scada_data = self._sim_generator.send(False)

            if self._scenario_sim.f_msx_in is not None:
                cur_time = int(scada_data.sensor_readings_time[0])
                cur_hyd_scada_data = self._hydraulic_scada_data.\
                    extract_time_window(cur_time, cur_time)
                scada_data.join(cur_hyd_scada_data)

            if self.autoreset is True:
                return scada_data
            else:
                return scada_data, False
        except StopIteration:
            if self.autoreset is True:
                _, info = self.reset()
                return info["scada_data"]
            else:
                return None, True

    def reset(self, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
              ) -> tuple[np.ndarray, dict]:
        """
        Resets this environment to an initial internal state, returning an
        initial observation and info.

        Parameters
        ----------
        seed : `int`, optional
            The seed that is used to initialize the environment's PRNG.

            The default is None.
        options : `dict[str, Any]`, optional
            Additional information to specify how the environment is reset
            (optional, depending on the specific environment).

            The default is None.

        Returns
        -------
        tuple[`numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, dict]
            Observation (`numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_),
            {"scada_data": `ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_}
            (`epyt_flow.simulation.ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_ as additional info).
        """
        Env.reset(self, seed=seed)

        if self._reload_scenario_when_reset is True:
            scada_data = super().reset()
        else:
            if self._scenario_sim is None:
                self._scenario_sim = ScenarioSimulator(scenario_config=self._scenario_config)
            else:
                # Abort current simulation if any is runing
                try:
                    next(self._sim_generator)
                    self._sim_generator.send(True)
                except StopIteration:
                    pass

            if self._scenario_sim.f_msx_in is not None:
                hyd_export = os.path.join(get_temp_folder(), f"epytflow_env_MSX_{uuid.uuid4()}.hyd")
                sim = self._scenario_sim.run_hydraulic_simulation
                self._hydraulic_scada_data = sim(hyd_export=hyd_export)

                gen = self._scenario_sim.run_advanced_quality_simulation_as_generator
                self._sim_generator = gen(hyd_export, support_abort=True)
            else:
                gen = self._scenario_sim.run_hydraulic_simulation_as_generator
                self._sim_generator = gen(support_abort=True)

            scada_data = self._next_sim_itr()

        r = scada_data
        if isinstance(r, tuple):
            r, _ = r
        r = self._get_observation(r)

        return r, {"scada_data": scada_data}

    def _get_observation(self, scada_data: ScadaData) -> np.ndarray:
        if scada_data is not None:
            return scada_data.get_data().flatten().astype(np.float32)
        else:
            return None

    @abstractmethod
    def _compute_reward_function(self, scada_data: ScadaData) -> float:
        """
        Computes the current reward based on the current sensors readings (i.e. SCADA data).

        Parameters
        ----------
        scada_data :`epyt_flow.simulation.ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_
            Current sensor readings.

        Returns
        -------
        `float`
            Current reward.
        """
        raise NotImplementedError()

    def step(self, action: np.ndarray) -> tuple[np.ndarray, float, bool, bool, dict]:
        """
        Performs the next step by applying an action and observing the next
        state together with a reward.

        Parameters
        ----------
        action : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Actions to be executed.

        Returns
        -------
        tuple[`numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, float, bool, bool, dict]
            Observation (`numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_), reward, terminated, False (truncated), {"scada_data": `ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_}
            (`epyt_flow.simulation.ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_ as additional info).
        """
        # Apply actions
        for action_value, action in zip(action, self._action_space):
            action.apply(self, action_value)

        # Run one simulation step and observe the sensor readings (SCADA data)
        if self.autoreset is False:
            current_scada_data, terminated = self._next_sim_itr()
        else:
            terminated = False
            current_scada_data = self._next_sim_itr()

        if isinstance(current_scada_data, tuple):
            current_scada_data, _ = current_scada_data

        if terminated is False:
            obs = self._get_observation(current_scada_data)

            # Calculate reward
            current_reward = self._compute_reward_function(deepcopy(current_scada_data))
        else:
            obs = None
            current_reward = None

        # Return observation and reward
        return obs, current_reward, terminated, False, {"scada_data": current_scada_data}



"""
This module contains a class for modeling pump speed action spaces.
"""
from typing import Optional
from epyt_flow.gym import ScenarioControlEnv
from gymnasium.spaces import Space, Box

from .actions import Action


class PumpSpeedAction(Action):
    """
    Action for controlling the speed of a pump.

    Parameters
    ----------
    pump_id : `str`
        ID of the pump.
    speed_lower_bound : `float`, optional
        Lower bound on the pump speed.

        The default is zero.
    speed_upper_bound : `float`, optional
        Upper bound on the pump speed.
        If None, no upper bound restriction for the agent.

        The default is None.
    """
    def __init__(self, pump_id: str, speed_lower_bound: float = 0.,
                 speed_upper_bound: Optional[float] = None, **kwds):
        if not isinstance(pump_id, str):
            raise TypeError("'pump_id' must be an instance of 'str' " +
                            f"but not of '{type(pump_id)}'")
        if not isinstance(speed_lower_bound, float):
            raise TypeError("'speed_lower_bound' must be an instance of 'float' " +
                            f"but not of '{type(speed_lower_bound)}'")
        if speed_lower_bound < 0:
            raise ValueError("'speed_lower_bound' can not be negative!")
        if speed_upper_bound is not None:
            if not isinstance(speed_upper_bound, float):
                raise TypeError("'speed_upper_bound' must be an instance of 'float' " +
                                f"but not of '{type(speed_upper_bound)}'")
            if speed_upper_bound < 0:
                raise ValueError("'speed_upper_bound' can not be negative!")

        self._pump_id = pump_id
        self._speed_lower_bound = speed_lower_bound
        self._speed_upper_bound = speed_upper_bound \
            if speed_upper_bound is not None else float("inf")

        super().__init__(**kwds)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._pump_id == other.pump_id and \
            self._speed_lower_bound == other.speed_lower_bound and\
            self._speed_upper_bound == other.speed_upper_bound

    def __str__(self) -> str:
        return super().__str__() + f"Pump ID: {self._pump_id} " +\
            f"Speed lower bound: {self._speed_lower_bound} " +\
            f"Speed upper bound: {self._speed_upper_bound}"

    @property
    def pump_id(self) -> str:
        """
        Return the ID of the pump.

        Returns
        -------
        `str`
            ID of the pump.
        """
        return self._pump_id

    @property
    def speed_lower_bound(self) -> float:
        """
        Returns the lower bound of the pump speed that can be set by the agent.

        Returns
        -------
        `float`
            Lower bound on the pump speed.
        """
        return self._speed_lower_bound

    @property
    def speed_upper_bound(self) -> float:
        """
        Returns the upper bound of the pump speed that can be set by the agent.

        Returns
        -------
        `float`
            Upper bound on the pump speed.
        """
        return self._speed_upper_bound

    def to_gym_action_space(self) -> Space:
        return Box(low=self._speed_lower_bound, high=self._speed_upper_bound)

    def apply(self, env: ScenarioControlEnv, action_value: float) -> None:
        env.set_pump_speed(self._pump_id, action_value)



"""
This module provides functions for evaluating policies/agents/control strategies on environments.
"""
from typing import Callable, Union
import numpy as np
from epyt_flow.simulation import ScadaData
from gymnasium import Wrapper

from .envs import RlEnv


def evaluate_policy(env: Union[RlEnv, Wrapper], policy: Callable[[np.ndarray], np.ndarray],
                    n_max_iter: int = 10) -> tuple[list[float], ScadaData]:
    """
    Evaluates a given policy/agent/control strategy for a given environment --
    i.e. the policy/agent is applied to the environment and the rewards and
    `ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_
    observations over time are recorded.

    Parameters
    ----------
    env : :class:`~epyt_control.envs.rl_env.RlEnv` or `gymnasium.Wrapper <https://gymnasium.farama.org/api/wrappers/#gymnasium.Wrapper>`_
        The environment.

        Note that in the case of a
        `gymnasium.Wrapper <https://gymnasium.farama.org/api/wrappers/#gymnasium.Wrapper>`_
        instance, the underlying environment must be an instance of
        :class:`~epyt_control.envs.rl_env.RlEnv`.
    policy : `Callable[[numpy.ndarray], numpy.ndarray]`
        Policy/Agent/Control strategy to be evaluated.
    n_max_iter : `int`, optional
        Upper bound on the number of iterations that is used for evaluating the given policy/agent.

        The default is 1.

    Returns
    -------
    tuple[list[float], `epyt_flow.simulation.ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_]
        Tuple of rewards over time and a
        `epyt_flow.simulation.ScadaData <https://epyt-flow.readthedocs.io/en/stable/epyt_flow.simulation.scada.html#epyt_flow.simulation.scada.scada_data.ScadaData>`_
        instance containing the WDN states over time.
    """
    if not isinstance(env, RlEnv) and not isinstance(env, Wrapper):
        raise TypeError("'env' must be an instance of 'epyt_control.envs.RlEnv' or " +
                        f"'gymnasium.Wrapper' but not of '{type(env)}'")
    if isinstance(env, Wrapper):
        if not isinstance(env.env, RlEnv):
            raise TypeError("The wrapped environment must be an insance of " +
                            f"'epyt_control.envs.RlEnv' but not of '{type(env.env)}'")
    if not callable(policy):
        raise TypeError("'policy' must be callable -- " +
                        "i.e. mapping observations (numpy.ndarray) to actions (numpy.ndarray)")
    if not isinstance(n_max_iter, int) or n_max_iter < 1:
        raise ValueError("'n_max_iter' must be an integer >= 1")

    rewards = []
    scada_data = None

    obs, _ = env.reset()
    for _ in range(n_max_iter):
        action = policy(obs)
        obs, reward, terminated, _, info = env.step(action)
        if terminated is True:
            break

        rewards.append(reward)
        current_scada_data = info["scada_data"]
        if scada_data is None:
            scada_data = current_scada_data
        else:
            scada_data.concatenate(current_scada_data)

    env.close()

    return rewards, scada_data
